/*
**  Subsystem:  WU-FTPD FTP Server
**  Purpose:    Change WU-FTPD Guest Passwords
**  File Name:  privatepw.c               
**                                                        
**  usage: privatepw [-c] [-f passwordfile] [-g group] accessgroup
**         privatepw [-d] [-f passwordfile] accessgroup
**         privatepw [-l] [-f passwordfile] 
**         -c:           creates a new file.
**         -d:           deletes specified accessgroup.
**         -l:           list contents of ftpgroups file.
**         -f ftpgroups: updates the specified file.
**         -g group:     set real group to the specified group.
**
**  This software is Copyright (c) 1997 by Kent Landfield
**
**  Permission is hereby granted to copy, distribute or otherwise 
**  use any part of this package as long as you do not try to make 
**  money from it or pretend that you wrote it.  This copyright 
**  notice must be maintained in any copy made.  If you are interested
**  in commercial distribution of this package as a whole or part of
**  another commercial package, contact the author for permission.
**  If you are a hardware or operating system vendor and would like 
**  loan equipment to the author in order to have this ported to your 
**  platform, please contact the author so it can be scheduled.
**
**  Use of this software constitutes acceptance for use in an AS IS 
**  condition. There are NO warranties with regard to this software.  
**  In no event shall the author be liable for any damages whatsoever 
**  arising out of or in connection with the use or performance of this 
**  software.  Any use of this software is at the user's own risk.
**
**  If you make modifications to this software that you feel 
**  increases it usefulness for the rest of the community, please 
**  email the changes, enhancements, bug fixes as well as any and 
**  all ideas to me. This software is going to be maintained and 
**  enhanced as deemed necessary by the community.
**                  
**                  Kent Landfield
**                  kent@landfield.com
*/

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <grp.h>
#include <unistd.h>
#include "pathnames.h"

#define BUFLEN 256
#define GROUPLEN 8

char *tmp;
char line[BUFLEN];

FILE *fp;

static unsigned char itoa64[] =         /* 0 ... 63 => ascii - 64 */
        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

void usage() 
{
   fprintf(stderr,"usage: privatepw [-c] [-f ftpgroups] [-g group] accessgroup\n");
   fprintf(stderr,"       privatepw [-d] [-f ftpgroups] accessgroup\n");
   fprintf(stderr,"       privatepw [-l] [-f ftpgroups]\n");
   fprintf(stderr,"\t\t-c:           creates a new file.\n");
   fprintf(stderr,"\t\t-d:           deletes specified accessgroup.\n");
   fprintf(stderr,"\t\t-l:           list contents of ftpgroups file.\n");
   fprintf(stderr,"\t\t-f ftpgroups: updates the specified file.\n");
   fprintf(stderr,"\t\t-g group:     set real group to the specified group.\n");
   exit(1);
}

void to64(s, v, n)
    register char *s;
    register long v;
    register int n;
{
    while (--n >= 0) {
        *s++ = itoa64[v&0x3f];
        v >>= 6;
    }
}

void terminate() 
{
    if (tmp)
        unlink(tmp);
    exit(1);
}

void catchintr() 
{
    fprintf(stderr,"Interrupted.\n");
    terminate();
}

char *savit(s)
    char *s;
{
    char *d;

    if ((d = (char *)malloc(strlen(s) + 1)) == NULL) {
        fprintf(stderr,"Whoa... Malloc failed.\n");
        terminate();
    }
    strcpy(d,s);
    return(d);
}

int confirmed(accessgroup)
char *accessgroup;
{
    register int ch;

    printf("Delete %s: Are your sure ? (y/n) ",accessgroup);
    ch = getc(stdin);
    if (ch == 'y')
        return(1);
    return(0);
}

char *getgroup(msg)
    char *msg;
{
    struct group *getgrnam();
    register int ch;
    register char *p;
    static char buf[GROUPLEN + 1];

    fputs(msg, stderr);
    rewind(stderr);            /* implied flush */
    for (p = buf; (ch = getc(stdin)) != EOF && ch != '\n';)
        if (p < buf + GROUPLEN)
            *p++ = ch;
    *p = '\0';

    if (getgrnam(buf) == NULL) {
        fprintf(stderr,"Invalid group \'%s\' specified\n",buf);
        terminate();
    }
    return(buf);
}

void addrecord(accessgroup, sysgroup, msg, f)
    char *accessgroup;
    char *sysgroup;
    char *msg;
    FILE *f;
{
    char *crypt();
    char *pw, *cpw, salt[3];

    printf("%s %s\n",msg,accessgroup);

    if (sysgroup[0] == NULL)
        strcpy(sysgroup,getgroup("Real System Group to use: "));

    pw = savit((char *) getpass("New password: "));
    if (strcmp(pw,(char *) getpass("Re-type new password: "))) {
        fprintf(stderr,"They don't match, sorry.\n");
        if (tmp)
            unlink(tmp);
        exit(1);
    }

    srand((int)time((time_t *)NULL));
    to64(&salt[0],rand(),2);
    cpw = crypt(pw,salt);
    free(pw);
    fprintf(f,"%s:%s:%s\n",accessgroup,cpw,sysgroup);
}

void list_privatefile(privatefile)
char *privatefile;
{
    if ((fp = fopen(privatefile,"r")) == NULL) {
        fprintf(stderr, "Could not open %s file.\n", privatefile);
        exit(1);
    }

    printf("\nWU-FTPD Private file: %s\n",privatefile);
    printf("accessgroup : password : system group\n");
    printf("-------\n");

    while(fgets(line, BUFLEN, fp) != NULL) 
        fputs(line,stdout);
    printf("-------\n");
}

void main(argc, argv) 
    int argc;
    char *argv[];
{
    int getopt();

    struct stat stbuf;

    extern char *optarg;
    extern int optind;
    extern int opterr;

    char realgroup[BUFLEN];
    char *passwdpath;
    char *cp;

    char accessgroup[BUFLEN];
    char w[BUFLEN];
    char command[BUFLEN];

    int create;
    int delete;
    int list;
    int found;
    int lineno;
    int c;

    FILE *tfp;

    opterr = 0;
    create = 0;
    delete = 0;
    list = 0;

    tmp = NULL;
    realgroup[0] = NULL;
    
    passwdpath = _PATH_PRIVATE;

    if (argc == 1) 
        usage();

    while ((c = getopt(argc, argv, "cdf:g:l")) != EOF) {
       switch (c) {
          case 'd':
              delete++;
              break;
          case 'c':
              create++;
              break;
          case 'f':
              passwdpath = optarg;
              break;
          case 'g':
              strcpy(realgroup,optarg);
              if (getgrnam(realgroup) == NULL) {
                  fprintf(stderr,"Invalid group \'%s\' specified\n",realgroup);
                  exit(1);
              }
              break;
          case 'l':
              list++;
              break;
          default:
              usage();
       }
    }

    if (list) {
        list_privatefile(passwdpath);
        exit(0);
    }

    if (optind >= argc) {
        fprintf(stderr,"Need to specify an accessgroup name.\n");
        usage();
    }
    
    signal(SIGINT,(void (*)())catchintr);

    strcpy(accessgroup,argv[optind]);

    if (create) {
        if (stat(passwdpath,&stbuf) == 0) {
            fprintf(stderr,"%s exists, cannot create it.\n", passwdpath);
            fprintf(stderr,"Remove -c option or use the -f option to specify another.\n");
            exit(1);
        }

        if ((tfp = fopen(passwdpath,"w")) == NULL) {
            fprintf(stderr,"Could not open \"%s\" for writing.\n",passwdpath);
            perror("fopen");
            exit(1);
        }
  
        tmp = passwdpath;

        printf("Creating WU-FTPD Private file: %s\n",passwdpath);
        addrecord(accessgroup,realgroup,"Adding accessgroup",tfp);

        fclose(tfp);
        exit(0);
    }

    tmp = tmpnam(NULL);

    if ((tfp = fopen(tmp,"w")) == NULL) {
        fprintf(stderr,"Could not open temp file.\n");
        exit(1);
    }

    if ((fp = fopen(passwdpath,"r")) == NULL) {
        fprintf(stderr, "Could not open %s file.\n", passwdpath);
        fprintf(stderr,"Use -c option to create new one.\n");
        exit(1);
    }

    lineno = 0;
    found = 0;

    while(fgets(line,BUFLEN,fp) != NULL) {
        lineno++;

        if (found || (line[0] == '#') || (!line[0])) {
            fputs(line,tfp);
            continue;
        }

        strcpy(w,line);

        if ((cp = strchr(w,':')) == NULL) {
             fprintf(stderr, "%s: line %d: invalid record format.\n",passwdpath,lineno);
             continue;
        }
        *cp++ = '\0';

        if ((cp = strchr(cp,':')) == NULL) {
             fprintf(stderr, "%s: line %d: invalid record format.\n",passwdpath,lineno);
             continue;
        }
        *cp++ = '\0';

        if (strcmp(accessgroup,w)) {
            fputs(line,tfp);
            continue;
        }
        else {
            if (delete) {
                if (!confirmed(accessgroup))
                    terminate();
            }
            else {
                if (realgroup[0] == NULL) {
                    strcpy(realgroup,cp);
                    if ((cp = strchr(realgroup,'\n')) != NULL)
                         *cp = '\0';
                }
                addrecord(accessgroup,realgroup,"Updating accessgroup",tfp);
            }
            found = 1;
        }
    }

    if (!found && !delete) 
        addrecord(accessgroup,realgroup,"Adding accessgroup",tfp);
    else if (!found && delete) {
        fprintf(stderr, "%s not found in %s.\n",accessgroup,passwdpath);
        terminate();
    }

    fclose(fp);
    fclose(tfp);

    sprintf(command,"cp %s %s",tmp,passwdpath);
    system(command);
    unlink(tmp);
}
